<?php

class Maarch_Sniffs_Functions_FunctionCallSniff 
    implements PHP_CodeSniffer_Sniff
{

    public function register()
    {
        return array(T_STRING);

    }//end register()


    /**
     * Processes this sniff, when one of its tokens is encountered.
     *
     * @param PHP_CodeSniffer_File $phpcsFile The current file being checked.
     * @param int $stackPtr  The position of the current token in the
     *                                        stack passed in $tokens.
     *
     * @return void
     */
    public function process(PHP_CodeSniffer_File $phpcsFile, $stackPtr)
    {
        
        $tokens = $phpcsFile->getTokens();

        // Find the next non-empty token.
        $funcStart = $phpcsFile->findNext(
            PHP_CodeSniffer_Tokens::$emptyTokens, 
            ($stackPtr + 1), null, true
        );

        if ($tokens[$funcStart]['code'] !== T_OPEN_PARENTHESIS) {
            // Not a function call.
            return;
        }

        if (isset($tokens[$funcStart]['parenthesis_closer']) === false) {
            // Not a function call.
            return;
        }

        // Find the previous non-empty token.
        $search   = PHP_CodeSniffer_Tokens::$emptyTokens;
        $search[] = T_BITWISE_AND;
        $previous = $phpcsFile->findPrevious($search, ($stackPtr - 1), 
                                             null, true);
        if ($tokens[$previous]['code'] === T_FUNCTION) {
            // It's a function definition, not a function call.
            return;
        }

        $funcEnd = $tokens[$funcStart]['parenthesis_closer'];
        $functionNamePtr = $stackPtr;
        
        //$keywordStart = $tokens[$stackPtr]['column'];
        $funcNestingLevel = array_key_exists('nested_parenthesis', 
                                           $tokens[$functionNamePtr])
                        ? (count($tokens[$functionNamePtr]['nested_parenthesis']) + 1)
                        : 1;
        $funcStartLine = $tokens[$funcStart]['line'];
        $funcEndLine = $tokens[$funcEnd]['line'];
        $isSingleLine = ($funcStartLine === $funcEndLine);
        //$isAssociative = (bool) $phpcsFile->findNext(array(T_DOUBLE_ARROW), 
        //                                             $arrayStart, $arrayEnd);
        $firstArg = $phpcsFile->findNext(
            array(T_WHITESPACE, T_COMMENT), 
            ($funcStart + 1), 
            ($funcEnd + 1), 
            true
        );
        $lastArg = $phpcsFile->findPrevious(
            array(T_WHITESPACE, T_COMMENT), 
            ($funcEnd - 1), 
            $funcStart, 
            true
        );
        

        if ($funcStart != ($stackPtr + 1)) {
            $error = 'There must be no space between the function name '
                   . 'and the opening parenthesis';
            $phpcsFile->addError($error, $funcStart - 1, 'SpaceAfterFunctionName');
        }

        // Check for empty arg list.
        if ($firstArg === $funcEnd) {
            // Empty array, but if the brackets aren't together, 
            // there's a problem.
            if (($funcEnd - $funcStart) !== 1) {
                $error = 'Empty function call must have no space ' 
                       . 'between the parentheses';
                $phpcsFile->addError($error, $firstArg - 1, 
                					 'SpaceInEmptyFunctionCall');

                // We can return here because there is nothing else to check. 
                // All code below can assume that the array is not empty.
                return;
            }
        }
        
        //check for spaces after commas
        $nextComma  = $funcStart;
        while ( 
            $nextComma = $phpcsFile->findNext(
                array(T_COMMA), 
                ($nextComma + 1), 
                $funcEnd
            )
        ) {
            $nextArg = $phpcsFile->findNext(
                array(T_WHITESPACE, T_COMMENT), 
                $nextComma + 1,
                $funcEnd,
                true
            );
            if ($tokens[$nextArg]['line'] !== $tokens[$nextComma]['line']) {
                $isEndOfLine = true;
            } else {
                $isEndOfLine = false;
            }
            
            if (!$isEndOfLine) {
                //check if there is one space after each comma
                if ($tokens[$nextComma + 1]['code'] !== T_WHITESPACE) {
                    $content = $tokens[($nextComma + 1)]['content'];
                    $error = 'Expected 1 space between comma and "%s"; 0 found';
                    $data = array($content);
                    $phpcsFile->addError($error, $nextComma, 
                                         'NoSpaceAfterComma', $data);
                } else {
                    $spaceLength = strlen($tokens[$nextComma + 1]['content']); 
                    if ($spaceLength !== 1) {
                        $content = $tokens[($nextComma + 2)]['content'];
                        $error = 'Expected 1 space between comma and "%s"; '
                               . '%s found';
                        $data = array($content, $spaceLength);
                        $phpcsFile->addError($error, $nextComma + 1, 
                                             'SpaceAfterComma', $data);
                    }
                }
            }
            
        }
        
        if ($isSingleLine) {
            //There must be no space before the closing parenthesis
            if ($tokens[$lastArg]['code'] === T_COMMA) {
                $error   = 'Last item must not be followed by a comma';
                $phpcsFile->addError($error, $lastArg, 
                    				 'NoCommaBeforeBracket');
                return;
            } elseif ($lastArg !== ($funcEnd - 1)) {
                $error   = 'There must be no space before closing parentheses';
                $phpcsFile->addError($error, $lastArg + 1, 
                                     'NoSpaceBeforeClosingParentheses');
            }
            //There must be no space after the opening parenthesis
            if ($firstArg !== ($funcStart + 1)) {
                $error   = 'There must be no space after opening parentheses';
                $phpcsFile->addError($error, $firstArg - 1, 
                                     'NoSpaceAfterOpeningParentheses');
            }
        
            // end singleline checks, start multiline  
        } else {
            
            $startUnderFunc = ($tokens[$firstArg]['line'] !== $funcStartLine);
            $parenOnNewLine = ($funcEndLine !== $tokens[$lastArg]['line']);
            
            if ($startUnderFunc) {
                if (!$parenOnNewLine) {
                    $error = 'Closing parenthesis of function call must be '
                           . 'on a new line';
                    $phpcsFile->addError($error, $funcEnd, 
                                     'CloseBraceNewLine');
                }
            } else {
                if ($parenOnNewLine) {
                    // The closing parenthesis must be on the last line
                    $error = 'The closing parenthesis must be on the '
                           . 'last line.';
                    $phpcsFile->addError($error, $funcEnd, 
                                         'CloseBraceNewLine');
                }
            }
            
            
            //
            // Indentation check
            //
            if ($startUnderFunc) {
                $firstItemOnCallLine = $phpcsFile->findFirstOnLine(
                    array(T_WHITESPACE), 
                    $funcStart, 
                    true
                );
                $expectedIndent = $tokens[$firstItemOnCallLine]['column'] + 4;
            } else {
                $expectedIndent = $tokens[$firstArg]['column'];
            }
            
            $currentLine = $tokens[$funcStart]['line'];
            $currentTokenPtr = $funcStart;
            while (
                $currentTokenPtr = $phpcsFile->findNext(
                    array(T_WHITESPACE), 
                    $currentTokenPtr + 1, 
                    $funcEnd, 
                    true
                )
            ) {
                $currentToken = $tokens[$currentTokenPtr];
                //if it is an array in an array, we pass checks
                $curNestingLevel = count($currentToken['nested_parenthesis']);
                if ($curNestingLevel !== $funcNestingLevel) {
                    $currentLine = $currentToken['line'];
                    continue;
                }
                
                // To verify indentation, we only care about first 
                // token on the line
                if ($currentToken['line'] === $currentLine) {
                    continue;
                }
                
                // here is the actual check
                $currentLine = $currentToken['line'];
                if ($currentToken['column'] !== $expectedIndent) {
                    $data = array(
                        $expectedIndent - 1, 
                        $currentToken['column'] - 1,
                    );
                    $error = 'Bad indent; expected %s, found %s.';
                    $phpcsFile->addError($error, $currentTokenPtr, 
                                         "BadIndent", $data);
                }
            }
        }
    }//end process()


}//end class

